/*
 *  matqConv.cpp
 *  fp
 *
 *  Created by Robert Delaney on 5/28/13.
 *  Copyright 2013 Bob Delaney's Science Software. All rights reserved.
 *
 */

#include "matq.h"
#include "matqConv.h"
#include "bfConv.h"
#include "stringUtils.h"

// fills matrix with bf rationals
// in xString row elements seperated by commas and rows seperated by semicolons
bool matqConvFromStr(matq& z, const char *xString)
{
	bf			**xt;
	INT32		i, j, k;
	char		*rowString, *numString;
	char		*xtString, ch;
	INT32		length;
	
	length = strlen(xString);
	
	if(!xString || !length)
	{
		z.array = NULL;
		return false;
	}
		
	xtString = (char*)malloc((length+1)*sizeof(char));
	xtString[0] = 0;
	
	strcpy(xtString, xString);
	
	// trim non-numerics from end of xtString
	i = length - 1;
	ch = xtString[i];
	while((ch<'0' || ch>'9') && i>0)
	{
		ch = xtString[--i];
	}
	xtString[i+1] = 0;
	
	z.nr = CountFields(xtString, ";");
	
	rowString = NthField(xtString, ";", 1);
	
	z.nc = CountFields(rowString, ",");
	
	xt = (bf**)malloc(z.nr*sizeof(bf*));
	for(i=0;i<z.nr;i++)
		xt[i] = (bf*)malloc(z.nc*sizeof(bf));
	
	for(i=0;i<z.nr;i++)
	{
		for(j=0;j<z.nc;j++)
		{
			numString = NthField(rowString, ",", j+1);
			if(!bfConvFromStr(xt[i][j], numString))
			{
				for(k=0;k<z.nr;k++)
					free(xt[k]);
				free(xt);
				z.array = NULL;
				return false;
			}
			free(numString);
		}
		
		free(rowString);
		if(i<(z.nr-1))
			rowString = NthField(xtString, ";", i+2);
	}
	
	z.array = (bf**)malloc(z.nr*sizeof(bf*));
	for(i=0;i<z.nr;i++)
		z.array[i] = (bf*)malloc(z.nc*sizeof(bf));
	
	for(i=0;i<z.nr;i++)
		for(j=0;j<z.nc;j++)
			z.array[i][j] = xt[i][j];
	
	for(k=0;k<z.nr;k++)
		free(xt[k]);
	free(xt);
	
	free(xtString);
	
	return true;
	
}/* matqConvFromStr */


void matqConvToStr(char*& matqStr, const matq& x)
{
	INT32		i, j;
	char		*numStr;
	INT32		allNumLen;
	
	allNumLen = 0;
	
	for(i=0;i<x.nr;i++)
		for(j=0;j<x.nc;j++)
		{
			bfConvToStr(numStr, x.array[i][j]);
			allNumLen = allNumLen + strlen(numStr);
			free(numStr);
		}
	// account for commas, spaces, and semicolons	
	allNumLen = allNumLen + 2*x.nr*x.nc + x.nr + 20;
	matqStr = (char*)malloc(allNumLen*sizeof(char));
	matqStr[0] = 0;
	
	for(i=0;i<x.nr;i++)
	{
		for(j=0;j<x.nc;j++)
		{
			bfConvToStr(numStr, x.array[i][j]);
			strcat(matqStr, numStr);
			if(j<(x.nc-1))
				strcat(matqStr, ", ");
			free(numStr);
		}
		if(i<(x.nr-1))
			strcat(matqStr, ";");
	}
	
}/* matqConvToStr */


